package com.twmacinta.net;

import java.io.*;
import java.net.*;

/**
 * Copyright (c) 2001 by Timothy W Macinta                <br>
 * http://www.twmacinta.com/                              <br>
 * http://www.twmacinta.com/myjava/ucon_timeout.php       <br>
 * <p>
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or (at your option) any later version.
 * <p>
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 * <p>
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 * @author Tim Macinta (twm@alum.mit.edu)
 **/

public class URLConnectionUtils {

  private static boolean native_methods_loaded = false;

  static {
    
    // load native method support

    try {
      File lib;

      // try package directory

      lib = new File(new File(new File(new File("com"), "twmacinta"), "net"), "URLConnectionUtils.so");

      // try main directory

      if (!lib.exists()) lib = new File("URLConnectionUtils.so");

      if (lib.exists()) {
	System.load(lib.getAbsolutePath());
      } else {

	// when all else fails, let the system hunt for it

	System.loadLibrary("URLConnectionUtils");
      }
      native_methods_loaded = true;
    } catch (UnsatisfiedLinkError e) {
      System.err.println("Warning:  Unable to load native methods for com.twmacinta.net.URLConnectionUtils.  This may make the system unable to terminate certain timed-out connections.");
    }
  }
  
  /**
   * This method provides a workaround for a bug in Sun's
   * HttpURLConnection.disconnect() method.  Sun's method
   * does not appropriately disconnect in a reliable manner.
   * All call to that method may not appropriately release
   * callers in other threads to such methods as getInputStream()
   * because the HttpURLConnection will attempt to reconnect if
   * no input has been received yet.  This essentially leaves
   * no (clean) way to disconnect a URLConnection.
   * <p>
   * This method provides a means for bypassing the reconnection
   * mechanism so that callers to URLConnection methods in other
   * threads which block before input is received will be
   * released through the throwing of an IOException.  This is
   * done in a very messy way, but there is unfortunately no
   * obvious, clean alternative which will allow the use of Sun's
   * existing network code.  A native method is used to bypass
   * Java's security restrictions and mess with the internal
   * fields of the HttpURLConnection so that it appears that it
   * has already made a failed reconnection attempt.
   * <p>
   * Note that the exception which is thrown in a thread blocked
   * waiting for input may be misleading.  It is guaranteed to
   * be an IOException, but no guarantees are made beyond that.
   * Presently, the internals of the HttpURLConnection are
   * manipulated such that doInput and doOutput are both false
   * and this may be reflected in the thrown exception.
   **/
  public static void disconnect(URLConnection ucon) {
    if (ucon == null) return;
    if (ucon instanceof HttpURLConnection) {
      HttpURLConnection http_ucon = (HttpURLConnection) ucon;
      if (native_methods_loaded) {
	_clobberHTTPUConInternals(http_ucon);
      }
      http_ucon.disconnect();
    } else {
      try {
	if (ucon.getDoInput()) ucon.getInputStream().close();
	else if (ucon.getDoOutput()) ucon.getOutputStream().close();
      } catch (IOException e) {}
    }
  }

  private static native void _clobberHTTPUConInternals(HttpURLConnection http_ucon);

}
